# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandera as pa
from pandera.typing import Series

# Ensure check method is registered or else would get SchemaInitError
# noreorder
from kolena._utils.dataframes.validators import _validate_locator  # noqa: F401


class LocatorEmbeddingsDataFrameSchema(pa.SchemaModel):
    key: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    """Unique key corresponding to model used for embeddings extraction. This is typically a locator."""

    locator: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    """External locator pointing to a sample in bucket."""

    embedding: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    Embedding vector (base64-encoded string of `np.ndarray`) corresponding to a searchable representation of the sample.
    """


class DatasetEmbeddingsDataFrameSchema(pa.SchemaModel):
    key: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    """Unique key corresponding to model used for embeddings extraction. This is typically a locator."""

    datapoint_id_object: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    String representation of the serialized datapoint id object from the dataset's id fields.
    """

    embedding: Series[pa.typing.String] = pa.Field(coerce=True)
    """
    Embedding vector (base64-encoded string of `np.ndarray`) corresponding to a searchable representation of the
        datapoint.
    """



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module surfaces plot definitions to visualize test-case-level data. [Evaluator](evaluator.md) implementations can
optionally compute plots using these definitions for visualization on the
[:kolena-results-16: Results](https://app.kolena.com/redirect/results) page.

The following plot types are available:

- [`CurvePlot`][kolena.workflow.CurvePlot]
- [`Histogram`][kolena.workflow.Histogram]
- [`BarPlot`][kolena.workflow.BarPlot]
- [`ConfusionMatrix`][kolena.workflow.ConfusionMatrix]
"""
from abc import ABCMeta
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
from pydantic.dataclasses import dataclass

from kolena._utils.datatypes import DataCategory
from kolena._utils.datatypes import DataObject
from kolena._utils.datatypes import DataType
from kolena._utils.datatypes import TypedDataObject
from kolena._utils.validators import ValidatorConfig

NumberSeries = Sequence[Union[float, int]]
"""A sequence of numeric values."""

NullableNumberSeries = Sequence[Union[float, int, None]]
"""A sequence of numeric values or `None`."""


class _PlotType(DataType):
    CURVE = "CURVE"
    CONFUSION_MATRIX = "CONFUSION_MATRIX"
    HISTOGRAM = "HISTOGRAM"
    BAR = "BAR"

    @staticmethod
    def _data_category() -> DataCategory:
        return DataCategory.PLOT


@dataclass(frozen=True, config=ValidatorConfig)
class AxisConfig(DataObject):
    """Configuration for the format of a given axis on a plot."""

    type: Literal["linear", "log"]
    """Type of axis to display. Supported options are `linear` and `log`."""


@dataclass(frozen=True, config=ValidatorConfig)
class Plot(TypedDataObject[_PlotType], metaclass=ABCMeta):
    """A data visualization shown when exploring model results in the web platform."""


@dataclass(frozen=True, config=ValidatorConfig)
class Curve(DataObject):
    """A single series on a [`CurvePlot`][kolena.workflow.CurvePlot]."""

    x: NumberSeries
    """The `x` coordinates of this curve. Length must match the provided `y` coordinates."""

    y: NumberSeries
    """The `y` coordinates of this curve. Length must match the provided `x` coordinates."""

    label: Optional[str] = None
    """
    Optionally specify an additional label (in addition to the associated test case) to apply to this curve, for use
    when e.g. there are multiple curves generated per test case.
    """

    extra: Optional[Dict[str, NumberSeries]] = None
    """
    Optionally specify additional series shown when hovering over the plot. For example, when plotting a
    precision-recall curve, it is desirable to include an extra series `threshold` to specify the confidence threshold
    value at which a given precision-recall point occurs.
    """

    def __post_init__(self) -> None:
        if len(self.x) != len(self.y):
            raise ValueError(
                f"Series 'x' (length: {len(self.x)}) and 'y' (length: {len(self.y)}) have different lengths",
            )
        for key, series in (self.extra or {}).items():
            if len(series) != len(self.x):
                raise ValueError(
                    f"Series '{key}' (length: {len(series)}) must match length of 'x' and 'y' (length: {len(self.x)})",
                )


@dataclass(frozen=True, config=ValidatorConfig)
class CurvePlot(Plot):
    """
    A plot visualizing one or more curves per test case.

    Examples include Receiver Operating Characteristic (ROC) curves, Precision versus Recall (PR) curves,
    Detection-Error Tradeoff (DET) curves, etc.
    """

    title: str
    """The title for the plot."""

    x_label: str
    """The label describing the plot's `x` axis."""

    y_label: str
    """The label describing the plot's `y` axis."""

    curves: List[Curve]
    """
    A test case may generate zero or more curves on a given plot. However, under most circumstances, a single curve
    per test case is desirable.
    """

    x_config: Optional[AxisConfig] = None
    """
    Custom options to allow for control over the display of the plot `x` axis. See
    [`AxisConfig`][kolena.workflow.AxisConfig] for details.
    """

    y_config: Optional[AxisConfig] = None
    """
    Custom options to allow for control over the display of the plot `y` axis. See
    [`AxisConfig`][kolena.workflow.AxisConfig] for details.
    """

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.CURVE


@dataclass(frozen=True, config=ValidatorConfig)
class Histogram(Plot):
    """
    A plot visualizing distribution of one or more continuous values, e.g. distribution of an error metric across all
    samples within a test case.

    For visualization of discrete values, see [`BarPlot`][kolena.workflow.BarPlot].
    """

    title: str
    """The title for the plot."""

    x_label: str
    """The label describing the plot's `x` axis."""

    y_label: str
    """The label describing the plot's `y` axis."""

    buckets: NumberSeries
    """
    A Histogram requires intervals to bucket the data. For `n` buckets, `n+1` consecutive bounds must be specified in
    increasing order.
    """

    frequency: Union[NumberSeries, Sequence[NumberSeries]]
    """
    For `n` buckets, there are `n` frequencies corresponding to the height of each bucket. The frequency at index `i`
    corresponds to the bucket with bounds (`i`, `i+1`) in `buckets`.

    To specify multiple distributions for a given test case, multiple frequency series can be provided, corresponding
    e.g. to the distribution for a given class within a test case, with name specified in `labels`.

    Specify a list of labels corresponding to the different `frequency` series when multiple series are provided.
    Can be omitted when a single `frequency` series is provided.
    """

    labels: Optional[List[str]] = None
    """Specify the label corresponding to a given distribution when multiple are specified in `frequency`."""

    x_config: Optional[AxisConfig] = None
    """
    Custom options to allow for control over the display of the plot `x` axis. See
    [`AxisConfig`][kolena.workflow.AxisConfig] for details.
    """

    y_config: Optional[AxisConfig] = None
    """
    Custom options to allow for control over the display of the plot `y` axis. See
    [`AxisConfig`][kolena.workflow.AxisConfig] for details.
    """

    def __post_init__(self) -> None:
        n_buckets = len(self.buckets)
        if n_buckets < 2:
            raise ValueError(f"Minimum 2 entries required in 'buckets' series (length: {n_buckets})")
        buckets_arr = np.array(self.buckets)
        if not np.all(buckets_arr[1:] > buckets_arr[:-1]):  # validate strictly increasing
            raise ValueError("Invalid 'buckets' series: series must be strictly increasing")

        frequency_arr = np.array(self.frequency)
        actual_frequency_shape = np.shape(frequency_arr)
        if len(actual_frequency_shape) > 1 and self.labels is None:
            raise ValueError("'labels' are required when multiple 'frequency' series are provided")

        n_labels = len(self.labels or [])
        expected_frequency_shape = (n_labels, n_buckets - 1) if n_labels > 0 else (n_buckets - 1,)
        if actual_frequency_shape != expected_frequency_shape:
            raise ValueError(f"Invalid 'frequency' shape {actual_frequency_shape}: expected {expected_frequency_shape}")

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.HISTOGRAM


@dataclass(frozen=True, config=ValidatorConfig)
class BarPlot(Plot):
    """A plot visualizing a set of bars per test case."""

    title: str
    """The plot title."""

    x_label: str
    """Axis label for the axis along which the bars are laid out (`labels`)."""

    y_label: str
    """Axis label for the axis corresponding to bar height (`values`)."""

    labels: Sequence[Union[str, int, float]]
    """Labels for each bar with corresponding height specified in `values`."""

    values: NullableNumberSeries
    """Values for each bar with corresponding label specified in `labels`."""

    config: Optional[AxisConfig] = None
    """Custom format options to allow for control over the display of the numerical plot axis (`values`)."""

    def __post_init__(self) -> None:
        n_labels, n_values = len(self.labels), len(self.values)
        if n_labels == 0 or n_values == 0 or n_labels != n_values:
            raise ValueError(
                f"Series 'labels' (length: {n_labels}) and 'values' (length: {n_values}) "
                "must be equal length and non-empty",
            )

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.BAR


@dataclass(frozen=True, config=ValidatorConfig)
class ConfusionMatrix(Plot):
    """
    A confusion matrix. Example:

    ```python
    ConfusionMatrix(
        title="Cat and Dog Confusion",
        labels=["Cat", "Dog"],
        matrix=[[90, 10], [5, 95]],
    )
    ```

    Yields a confusion matrix of the form:

    ```
                Predicted

                Cat   Dog
               +----+----+
           Cat | 90 | 10 |
    Actual     +----+----+
           Dog |  5 | 95 |
               +----+----+
    ```
    """

    title: str
    """The plot title."""

    labels: List[str]
    """The labels corresponding to each entry in the square `matrix`."""

    matrix: Sequence[NullableNumberSeries]
    """A square matrix, typically representing the number of matches between class `i` and class `j`."""

    x_label: str = "Predicted"
    """The label for the `x` axis of the confusion matrix."""

    y_label: str = "Actual"
    """The label for the `y` axis of the confusion matrix."""

    def __post_init__(self) -> None:
        n_labels = len(self.labels)
        if n_labels < 2:
            raise ValueError(f"At least two labels required: got {n_labels}")
        if len(self.matrix) != n_labels:
            raise ValueError(f"Invalid number of matrix rows: got {len(self.matrix)}, expected {n_labels}")
        for i, row in enumerate(self.matrix):
            if len(row) != n_labels:
                raise ValueError(f"Invalid number of matrix columns in row {i}: {len(row)}, expected {n_labels}")

    @staticmethod
    def _data_type() -> _PlotType:
        return _PlotType.CONFUSION_MATRIX



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import List
from typing import Union

import pandas as pd

from kolena._api.v1.event import EventAPI
from kolena._api.v2.quality_standard import Path
from kolena._utils import krequests_v2 as krequests
from kolena._utils import log
from kolena._utils.instrumentation import with_event


def _format_quality_standard_result_df(quality_standard_result: dict) -> pd.DataFrame:
    df = pd.DataFrame.from_dict(quality_standard_result).rename_axis("stratification")

    df = df.explode("results_by_stratification")
    df = pd.concat(
        [df.drop(["results_by_stratification"], axis=1), df["results_by_stratification"].apply(pd.Series)],
        axis=1,
    )
    df.set_index(["test_case"], append=True, drop=True, inplace=True)

    df = df.explode("model_results")
    df["model"] = df["model_results"].apply(lambda x: x["model"]["name"])
    df["eval_config"] = df["model_results"].apply(lambda x: json.dumps(x["model"]["eval_config"]))
    df["metric_group"] = df["model_results"].apply(lambda x: list(x["results"].keys()))

    df = df.explode("metric_group")
    df["metric"] = df.apply(lambda x: list(x["model_results"]["results"][x["metric_group"]].keys()), axis=1)

    df = df.explode("metric")
    df["value"] = df.apply(lambda x: x["model_results"]["results"][x["metric_group"]][x["metric"]], axis=1)

    return df.pivot(columns=["model", "eval_config", "metric_group", "metric"], values="value")


@with_event(event_name=EventAPI.Event.FETCH_QUALITY_STANDARD_RESULT)
def download_quality_standard_result(
    dataset: str,
    models: List[str],
    metric_groups: Union[List[str], None] = None,
) -> pd.DataFrame:
    """
    Download quality standard result given a dataset and list of models.

    :param dataset: The name of the dataset.
    :param models: The names of the models.
    :param metric_groups: The names of the metric groups to include in the result.
    Note all metric groups are included when this value is `None`.
    :return: A Dataframe containing the quality standard result.
    """
    model_log = ", ".join([f"'{model}'" for model in models])
    log.info(f"downloading quality standard results for model(s) {model_log} on dataset '{dataset}'")
    response = krequests.get(
        Path.RESULT,
        params=dict(dataset_name=dataset, models=models, metric_groups=metric_groups),
        api_version="v2",
    )
    krequests.raise_for_status(response)
    return _format_quality_standard_result_df(response.json())



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABCMeta
from collections import defaultdict
from contextlib import contextmanager
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type

from pydantic import validate_call

from kolena._api.v1.core import TestSuite as CoreAPI
from kolena._api.v1.event import EventAPI
from kolena._api.v1.generic import TestSuite as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.consts import BatchSize
from kolena._utils.consts import FieldName
from kolena._utils.endpoints import get_test_suite_url
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import with_event
from kolena._utils.serde import from_dict
from kolena._utils.validators import validate_name
from kolena._utils.validators import ValidatorConfig
from kolena.errors import IncorrectUsageError
from kolena.errors import NotFoundError
from kolena.workflow import TestSample
from kolena.workflow._datatypes import TestSuiteTestSamplesDataFrame
from kolena.workflow._validators import assert_workflows_match
from kolena.workflow.test_case import TestCase
from kolena.workflow.test_sample import _METADATA_KEY
from kolena.workflow.workflow import Workflow


class TestSuite(Frozen, metaclass=ABCMeta):
    """
    A test suite groups together one or more [test cases][kolena.workflow.TestCase]. Typically a test suite represents a
    benchmark test dataset, with test cases representing different meaningful subsets, or slices, or this benchmark.

    Rather than importing this class directly, use the `TestSuite` type definition returned from
    [`define_workflow`][kolena.workflow.define_workflow.define_workflow].
    """

    workflow: Workflow
    """
    The workflow of this test suite. Automatically populated when constructing via test suite type returned from
    [`define_workflow`][kolena.workflow.define_workflow.define_workflow].
    """

    name: str
    """The unique name of this test suite."""

    version: int
    """
    The version of this test suite. A test suite's version is automatically incremented whenever it is edited via
    [`TestSuite.edit`][kolena.workflow.TestSuite.edit].
    """

    description: str
    """
    Free-form, human-readable description of this test suite. Can be edited at any time via
    [`TestSuite.edit`][kolena.workflow.TestSuite.edit].
    """

    test_cases: List[TestCase]
    """The [`TestCase`][kolena.workflow.TestCase] objects belonging to this test suite."""

    tags: Set[str]
    """The tags associated with this test suite."""

    _id: int
    _test_case_type: Type[TestCase]

    def __init_subclass__(cls) -> None:
        if not hasattr(cls, "workflow"):
            raise NotImplementedError(f"{cls.__name__} must implement class attribute 'workflow'")
        if not hasattr(cls, "_test_case_type"):
            raise NotImplementedError(f"{cls.__name__} must implement class attribute '_test_case_type'")
        super().__init_subclass__()

    @validate_call(config=ValidatorConfig)
    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        test_cases: Optional[List[TestCase]] = None,
        reset: bool = False,
        tags: Optional[Set[str]] = None,
    ):
        validate_name(name, FieldName.TEST_SUITE_NAME)
        self._validate_workflow()
        self._validate_test_cases(test_cases)

        try:
            other = self.load(name, version)
        except NotFoundError:
            other = self.create(name, description, test_cases, tags)
        self._populate_from_other(other)

        should_update_test_cases = test_cases is not None and test_cases != self.test_cases
        can_update_test_cases = reset or self.version == 0
        if should_update_test_cases and not can_update_test_cases:
            log.warn(f"reset=False, not updating test cases on test suite '{self.name}' (v{self.version})")
        to_update = [
            *(["test cases"] if should_update_test_cases and can_update_test_cases else []),
            *(["description"] if description is not None and description != self.description else []),
            *(["tags"] if tags is not None and tags != self.tags else []),
        ]
        if len(to_update) > 0:
            log.info(f"updating {', '.join(to_update)} on test suite '{self.name}' (v{self.version})")
            updated_test_cases = test_cases or self.test_cases if can_update_test_cases else self.test_cases
            self._hydrate(updated_test_cases, description=description, tags=tags)

        self._freeze()

    @classmethod
    def _validate_test_cases(cls, test_cases: Optional[List[TestCase]] = None) -> None:
        if test_cases:
            if any(cls.workflow != testcase.workflow for testcase in test_cases):
                raise TypeError("test case workflow does not match test suite's")

    @classmethod
    def _validate_workflow(cls) -> None:
        if cls == TestSuite:
            raise IncorrectUsageError("<TestSuite> must be subclassed. See `kolena.workflow.define_workflow`")

    def _populate_from_other(self, other: "TestSuite") -> None:
        with self._unfrozen():
            self._id = other._id
            self.name = other.name
            self.version = other.version
            self.description = other.description
            self.test_cases = other.test_cases
            self.tags = other.tags

    @classmethod
    def _create_from_data(cls, data: CoreAPI.EntityData) -> "TestSuite":
        assert_workflows_match(cls.workflow.name, data.workflow)
        obj = cls.__new__(cls)
        obj._id = data.id
        obj.name = data.name
        obj.version = data.version
        obj.description = data.description
        obj.test_cases = [cls._test_case_type._create_from_data(tc) for tc in data.test_cases]
        obj.tags = set(data.tags)
        obj._freeze()
        return obj

    def _hydrate(
        self,
        test_cases: List[TestCase],
        description: Optional[str] = None,
        tags: Optional[Set[str]] = None,
    ) -> None:
        with self.edit(reset=True) as editor:
            if description is not None:
                editor.description(description)
            for test_case in test_cases:
                editor.add(test_case)
            if tags is not None:
                editor.tags = tags

    @classmethod
    @with_event(event_name=EventAPI.Event.CREATE_TEST_SUITE)
    def create(
        cls,
        name: str,
        description: Optional[str] = None,
        test_cases: Optional[List[TestCase]] = None,
        tags: Optional[Set[str]] = None,
    ) -> "TestSuite":
        """
        Create a new test suite with the provided name.

        :param name: The name of the new test suite to create.
        :param description: Optional free-form description of the test suite to create.
        :param test_cases: Optionally specify a list of test cases to populate the test suite.
        :param tags: Optionally specify a set of tags to attach to the test suite.
        :return: The newly created test suite.
        """
        cls._validate_workflow()
        cls._validate_test_cases(test_cases)
        validate_name(name, FieldName.TEST_SUITE_NAME)
        request = CoreAPI.CreateRequest(
            name=name,
            description=description or "",
            workflow=cls.workflow.name,
            tags=list(tags) if tags is not None else None,
        )
        res = krequests.post(endpoint_path=API.Path.CREATE, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"created test suite '{name}' (v{obj.version}) ({get_test_suite_url(obj._id)})")
        if test_cases is not None:
            obj._hydrate(test_cases)
        return obj

    @classmethod
    @with_event(event_name=EventAPI.Event.LOAD_TEST_SUITE)
    def load(cls, name: str, version: Optional[int] = None) -> "TestSuite":
        """
        Load an existing test suite with the provided name.

        :param name: The name of the test suite to load.
        :param version: Optionally specify a particular version of the test suite to load. Defaults to the latest
            version when unset.
        :return: The loaded test suite.
        """
        cls._validate_workflow()
        request = CoreAPI.LoadByNameRequest(name=name, version=version)
        res = krequests.put(endpoint_path=API.Path.LOAD, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"loaded test suite '{name}' (v{obj.version}) ({get_test_suite_url(obj._id)})")
        return obj

    @classmethod
    @with_event(event_name=EventAPI.Event.LOAD_ALL_TEST_SUITES)
    def load_all(cls, *, tags: Optional[Set[str]] = None) -> List["TestSuite"]:
        """
        Load the latest version of all test suites with this workflow.

        :param tags: Optionally specify a set of tags to apply as a filter. The loaded test suites will include only
            test suites with tags matching each of these specified tags, i.e.
            `test_suite.tags.intersection(tags) == tags`.
        :return: The latest version of all test suites, with matching tags when specified.
        """
        cls._validate_workflow()
        request = CoreAPI.LoadAllRequest(workflow=cls.workflow.name, tags=list(tags) if tags is not None else None)
        res = krequests.put(endpoint_path=API.Path.LOAD_ALL.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=CoreAPI.LoadAllResponse, data=res.json())
        objs = [cls._create_from_data(test_suite) for test_suite in data.test_suites]
        tags_quoted = {f"'{t}'" for t in tags or {}}
        tags_message = f" with tag{'s' if len(tags) > 1 else ''} {', '.join(tags_quoted)}" if tags else ""
        log.info(f"loaded {len(objs)} '{cls.workflow.name}' test suites{tags_message}")
        return objs

    class Editor:
        _test_cases: List[TestCase]
        _reset: bool
        _description: str
        _initial_test_case_ids: List[int]
        _initial_description: str
        _initial_tags: Set[str]

        #: The tags associated with this test suite. Modify this list directly to edit this test suite's tags.
        tags: Set[str]

        @validate_call(config=ValidatorConfig)
        def __init__(self, test_cases: List[TestCase], description: str, tags: Set[str], reset: bool):
            self._test_cases = test_cases if not reset else []
            self._reset = reset
            self._description = description
            self._initial_test_case_ids = [tc._id for tc in test_cases]
            self._initial_description = description
            self._initial_tags = tags
            self.tags = tags

        @validate_call(config=ValidatorConfig)
        def description(self, description: str) -> None:
            """Update the description of the test suite."""
            self._description = description

        @validate_call(config=ValidatorConfig)
        def add(self, test_case: TestCase) -> None:
            """
            Add a test case to this test suite. If a different version of the test case already exists in this test
            suite, it is replaced.

            :param test_case: The test case to add to the test suite.
            """
            self._test_cases = [*(tc for tc in self._test_cases if tc.name != test_case.name), test_case]

        @validate_call(config=ValidatorConfig)
        def remove(self, test_case: TestCase) -> None:
            """
            Remove a test case from this test suite. Does nothing if the test case is not in the test suite.

            :param test_case: The test case to remove.
            """
            self._test_cases = [tc for tc in self._test_cases if tc.name != test_case.name]

        def _edited(self) -> bool:
            return (
                self._description != self._initial_description
                or self._initial_test_case_ids != [tc._id for tc in self._test_cases]
                or self._initial_tags != self.tags
            )

    @contextmanager
    @with_event(event_name=EventAPI.Event.EDIT_TEST_SUITE)
    def edit(self, reset: bool = False) -> Iterator[Editor]:
        """
        Edit this test suite in a context:

        ```python
        with test_suite.edit() as editor:
            # perform as many editing actions as desired
            editor.add(...)
            editor.remove(...)
        ```

        Changes are committed to the Kolena platform when the context is exited.

        :param reset: Clear all existing test cases in the test suite.
        """
        editor = self.Editor(self.test_cases, self.description, self.tags, reset)

        yield editor

        if not editor._edited():
            return

        log.info(f"editing test suite '{self.name}' (v{self.version})")
        request = CoreAPI.EditRequest(
            test_suite_id=self._id,
            current_version=self.version,
            name=self.name,
            description=editor._description,
            test_case_ids=[tc._id for tc in editor._test_cases],
            tags=list(editor.tags),
        )
        res = krequests.post(endpoint_path=API.Path.EDIT, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        test_suite_data = from_dict(data_class=CoreAPI.EntityData, data=res.json())
        self._populate_from_other(self._create_from_data(test_suite_data))
        log.success(f"edited test suite '{self.name}' (v{self.version}) ({get_test_suite_url(self._id)})")

    @with_event(event_name=EventAPI.Event.LOAD_TEST_SUITE_SAMPLES)
    def load_test_samples(self) -> List[Tuple[TestCase, List[TestSample]]]:
        """
        Load test samples for all test cases within this test suite.

        :return: A list of [`TestCase`s][kolena.workflow.TestCase], each paired with the list of
            [`TestSample`s][kolena.workflow.TestSample] it contains.
        """
        test_case_id_to_samples: Dict[int, List[TestSample]] = defaultdict(list)
        for df_batch in _BatchedLoader.iter_data(
            init_request=API.LoadTestSamplesRequest(
                test_suite_id=self._id,
                batch_size=BatchSize.LOAD_SAMPLES,
            ),
            endpoint_path=API.Path.INIT_LOAD_TEST_SAMPLES,
            df_class=TestSuiteTestSamplesDataFrame,
        ):
            for record in df_batch.itertuples():
                test_sample = self.workflow.test_sample_type._from_dict(
                    {**record.test_sample, _METADATA_KEY: record.test_sample_metadata},
                )
                test_case_id_to_samples[record.test_case_id].append(test_sample)

        test_case_id_to_test_case = {tc._id: tc for tc in self.test_cases}
        return [(test_case_id_to_test_case[tc_id], samples) for tc_id, samples in test_case_id_to_samples.items()]



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Dict
from typing import List
from typing import Optional

from pydantic import conint
from pydantic.dataclasses import dataclass

from kolena._api.v1.batched_load import BatchedLoad


class Path(str, Enum):
    REGISTER = "/dataset/register"
    LOAD_DATAPOINTS = "/dataset/load-datapoints"
    LOAD_DATASET = "/dataset/load-by-name"
    LIST_COMMITS = "/dataset/list-commits"


@dataclass(frozen=True)
class RegisterRequest:
    name: str
    id_fields: List[str]
    uuid: str
    sources: Optional[List[Dict[str, str]]]


@dataclass(frozen=True)
class LoadDatapointsRequest(BatchedLoad.BaseInitDownloadRequest):
    name: str
    commit: Optional[str] = None


@dataclass(frozen=True)
class LoadDatasetByNameRequest:
    name: str
    raise_error_if_not_found: bool = True


@dataclass(frozen=True)
class EntityData:
    id: int
    name: str
    description: str
    id_fields: List[str]


@dataclass(frozen=True)
class ListCommitHistoryRequest:
    name: str
    descending: bool = False
    offset: conint(strict=True, ge=0) = 0
    limit: conint(strict=True, ge=0, le=100) = 50


@dataclass(frozen=True)
class CommitData:
    commit: str
    timestamp: int
    user: str
    n_removed: int
    n_added: int


@dataclass(frozen=True)
class ListCommitHistoryResponse:
    records: List[CommitData]
    total_count: int
    descending: bool
    offset: int
    limit: int



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from collections.abc import Callable
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import pandas as pd


def df_apply(df: pd.DataFrame, func: Callable, **kwargs: Any) -> pd.DataFrame:
    """
    pandas.DataFrame.apply wrapper
    """
    default_kwargs = dict(convert_dtype=False)
    with warnings.catch_warnings():
        # Silence the warning for now -- pandas devs have deprecated this without providing a good alternative:
        # https://github.com/pandas-dev/pandas/pull/52257#issuecomment-1684888371
        warnings.simplefilter(action="ignore", category=FutureWarning)
        return df.apply(func, **default_kwargs, **kwargs)


def _flatten_dict(d: Dict[str, Any], parent_key: str = "", sep: str = ".", max_level: int = 0) -> Dict[str, Any]:
    items: List[Tuple] = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict) and max_level > 0:
            items.extend(_flatten_dict(v, new_key, sep, max_level - 1).items())
        else:
            items.append((new_key, v))
    return dict(items)


def _normalize_dicts(dict_list: List[Dict[str, Any]], sep: str = ".", max_level: int = 0) -> List[Dict[str, Any]]:
    # Flatten all dictionaries in the list
    flattened_dicts = [_flatten_dict(d, sep=sep, max_level=max_level) for d in dict_list]

    # Get all unique keys across all dictionaries
    all_keys = set().union(*flattened_dicts)

    # Normalize all dictionaries to have all keys, filling missing ones with None
    for d in flattened_dicts:
        for key in all_keys:
            if key not in d:
                d[key] = None

    return flattened_dicts


def json_normalize(data: List[Dict[str, Any]], max_level: int = 0) -> pd.DataFrame:
    """
    hand roll pandas.json_normalize implementation
    use None for missing value
    """
    normalized_data = _normalize_dicts(data, max_level=max_level)
    return pd.DataFrame.from_dict(normalized_data, dtype=object)
